import numpy as np
import pytest

from harmonia.reader.likelihoods import (
    chi_square,
    complex_normal_pdf,
    modified_student_pdf,
    multivariate_normal_pdf,
)


@pytest.mark.parametrize(
    "ndim,nsample",
    [
        (3, 10000),
    ]
)
def test_chi_square(ndim, nsample):

    samples = np.random.randn(nsample, ndim)
    covariance_matrix = np.eye(ndim)

    assert np.allclose(
        np.mean([
            chi_square(sample_vector, covariance_matrix)
            for sample_vector in samples
        ]),
        ndim,
        atol=10*np.sqrt(2 * ndim / nsample)  # allow 10-sigma deviation
    ), "Incorrect chi square evaluation."


@pytest.mark.parametrize(
    "data_vector,covariance_matrix,value",
    [
        ([1. + 1.j, - 1. - 1.j], np.eye(2), np.exp(-4) / np.pi ** 2),
    ]
)
def test_complex_normal_pdf(data_vector, covariance_matrix, value):
    print(value)
    assert np.isclose(
        complex_normal_pdf(data_vector, covariance_matrix, ret_log=False),
        value
    ), "Incorrect complex normal PDF evaluation."


@pytest.mark.parametrize(
    "data_vector,expectation_vector,covariance_matrix,degree,value",
    [
        ([1., -1.], [0., 0.], np.eye(2), 100, 0.0579565),
    ]
)
def test_modified_student_pdf(data_vector, expectation_vector,
                              covariance_matrix, degree, value):

    assert np.isclose(
        modified_student_pdf(
            data_vector, expectation_vector, covariance_matrix,
            degree, ret_log=False
        ), value
    ), "Incorrect modified Student PDF evaluation."


@pytest.mark.parametrize(
    "data_vector,expectation_vector,covariance_matrix,degree",
    [
        ([1., -1.], [0., 0.], np.eye(2), 1000),
    ]
)
def test_multivariate_normal_pdf(data_vector, expectation_vector,
                                 covariance_matrix, degree):

    assert np.isclose(
        multivariate_normal_pdf(
            data_vector, expectation_vector, covariance_matrix, ret_log=False
        ),
        modified_student_pdf(
            data_vector, expectation_vector, covariance_matrix,
            degree, ret_log=False
        ),
        rtol=1/np.sqrt(degree)
    ), (
        "Multivariat normal and modified Student PDF "
        "evaluations do not match in the large number limit."
    )



"""
###########################################################################
``Harmonia`` | Hybrid map inference from large-scale galaxy clustering
###########################################################################

``Harmonia`` is a Python package that combines spherical and Cartesian
Fourier maps and likelihoods for large-scale galaxy clustering analysis.

.. topic:: Licence Statement

    Copyright (C) 2020, M S Wang

    This program is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation, either version 3 of the License, or (at your
    option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program.  If not, see `<https://www.gnu.org/licenses/>`_.

"""
__author__ = "Mike S Wang"
__contact__ = "Mike S Wang"
__copyright__ = "Copyright 2020, Harmonia/M S Wang"
__date__ = "2020/06/01"
__description__ = "Hybrid map inference from large-scale galaxy clustering."
__email__ = "mike.wang@port.ac.uk"
__license__ = "GPLv3"
__url__ = "https://mikeswang.github.io/Harmonia/"
__version__ = "0.1.0"



import numpy as np
import pytest
from nbodykit.cosmology import Planck13, Planck15

from harmonia.cosmology.geometry import (
    differential_AP_distortion,
    redshift_from_distance,
)


@pytest.mark.parametrize(
    "z",
    [
        (1.0,),
        (2.0,),
        (3.0,),
    ]
)
def test_differential_AP_distortion(z):

    assert np.isclose(
        differential_AP_distortion(
            Planck13.comoving_distance, Planck15.comoving_distance
        )(z), 1.,
        rtol=1.e-3
    ), "Incorrect computation of differential AP distortion."


@pytest.mark.skip(reason="Unimplemented.")
def test_redshift_from_distance():
    print("Unimplemented test.")
    raise NotImplementedError



"""
Spectrum discretisation (:mod:`~harmonia.algorithms.discretisation`)
===========================================================================

Discretise the Fourier spectrum under boundary conditions.

.. autosummary::

    DiscreteSpectrum

|

"""
import logging

import numpy as np

from .bases import spherical_besselj, spherical_besselj_root


class DiscreteSpectrum:
    r"""Discrete Fourier spectrum under radial boundary conditions.

    The spectral modes are indexed by tuple :math:`(\ell, n)`, where
    :math:`\ell` is the spherical degree associated with the spherical
    harmonic and Bessel functions, and :math:`n` is the spherical depth
    associated with roots of the spherical Bessel function.

    The discrete wavenumbers are

    .. math::

        k_{\ell n} = \frac{u_{\ell n}}{R} \,,
        \quad \text{where} \quad
        \{ u_{\ell n}: n = 1, \dots, n_{\textrm{max},\ell} \}_{\ell}

    are zeros of the spherical Bessel functions of order :math:`\ell` if
    the boundary condition is Dirichlet, or zeros of their derivatives if
    the boundary condition is Neumann.  The maximum spherical depth
    :math:`n_{\textrm{max},\ell}` corresponds to the largest wavenumber
    allowed in the cutoff range.

    The normalisation coefficients derived from completeness relations are

    .. math::

        \kappa_{\ell n} =
            \begin{cases}
                (2/R^3) j_{\ell+1}^{-2}(u_{\ell n}) \,,
                \quad \text{for Dirichlet boundary conditions;} \\
                (2/R^3) j_{\ell}^{-2}(u_{\ell n}) \left[
                    1 - \ell (\ell + 1) / u_{\ell n}^2
                \right]^{-1} \,,
                \quad \text{for Neumann boundary conditions.}
            \end{cases}

    Parameters
    ----------
    radius : float
        Boundary radius.
    condition : {'dirichlet', 'neumann'}
        Either Dirichlet or Neumann boundary condition.
    highcut : float
        Upper cutoff wavenumber of the spectrum.
    maxdeg : int or None, optional
        Maximum spherical degree (default is `None`).
    lowcut : float, optional
        Lower cutoff wavenumber of the spectrum (default is 0.).
    mindeg : int, optional
        Minimum spherical degree (default is 0).
    comm : :class:`mpi4py.MPI.Comm`
        MPI communicator.

    Attributes
    ----------
    degrees : list of int
        Spherical degrees associated with the discrete spectrum.
    depths : list of int
        Spherical depths associated with each spherical degree.
    mode_counts : list of int
        Total number of spectral modes associated with each spherical
        degree counting spherical order multiplicities.
    roots : dict
        Spherical Bessel roots associated with the discrete spectrum as
        a dictionary accessed by integer spherical degrees.
    attrs : dict
        Discrete spectrum attributes including the following keys:
        'min_wavenumber' and 'max_wavenumber' for minimum and
        maximum wavenumbers, 'boundary_radius' and 'bounded_volume'
        for the boundary radius and bounded volume, and
        'boundary_condition' for the imposed boundary condition.

    """

    def __init__(self, radius, condition, highcut, maxdeg=None, lowcut=0.,
                 mindeg=0, comm=None):

        self.logger = logging.getLogger(self.__class__.__name__)
        self.comm = comm

        condition = self._alias(condition)

        self.attrs = {
            'min_wavenumber': lowcut,
            'max_wavenumber': highcut,
            'boundary_radius': radius,
            'bounded_volume': 4./3. * np.pi * radius ** 3,
            'boundary_condition': condition,
        }

        self._discretise(radius, condition, lowcut, highcut, mindeg, maxdeg)

        self._wavenumbers = None
        self._normalisations = None

    def __str__(self):

        str_info = "{}, boundary={}, {}<=wavenumber<={}".format(
            self.attrs['boundary_condition'],
            self.attrs['boundary_radius'],
            np.around(self.attrs['min_wavenumber'], decimals=4),
            np.around(self.attrs['max_wavenumber'], decimals=4),
        )

        return f"{self.__class__.__name__}({str_info})"

    def __getstate__(self):

        state = {
            attr: getattr(self, attr)
            for attr in [
                'attrs', 'degrees', 'depths', 'mode_counts',
                'roots', 'wavenumbers', 'normalisations'
            ]
        }

        return state

    def __setstate__(self, state):

        for attr, value in state.items():
            if attr in ['wavenumbers', 'normalisations']:
                setattr(self, '_' + attr, value)
            else:
                setattr(self, attr, value)

    @classmethod
    def _from_state(cls, state, comm=None):  # internal classmethod

        self = object.__new__(cls)

        self.__setstate__(state)

        self.logger = logging.getLogger(self.__class__.__name__)
        self.comm = comm

        return self

    @property
    def wavenumbers(self):
        r"""Discrete mode wavenumbers :math:`k_{\ell n}`.

        Returns
        -------
        dict
            Mode wavenumbers as a dictionary accessed by
            doublet tuples :math:`(\ell, n)`.

        """
        if self._wavenumbers is not None:
            return self._wavenumbers

        self._wavenumbers = {
            (ell, n_idx + 1) : u / self.attrs['boundary_radius']
            for ell in self.degrees
            for n_idx, u in enumerate(self.roots[ell])
        }

        if self.comm is None or self.comm.rank == 0:
            self.logger.debug("Spectral mode wavenumbers computed.")

        return self._wavenumbers

    @property
    def normalisations(self):
        r"""Normalisation coefficients :math:`\kappa_{\ell n}`.

        Returns
        -------
        dict
            Normalisation coefficients as a dictionary accessed by
            doublet tuples :math:`(\ell, n)`.

        """
        if self._normalisations is not None:
            return self._normalisations

        # pylint: disable=bad-continuation
        if self.attrs['boundary_condition'] == 'dirichlet':
            self._normalisations = {
                (ell, n_idx + 1): 2.
                    / self.attrs['boundary_radius'] ** 3
                    / spherical_besselj(ell + 1, u) ** 2
                for ell in self.degrees
                for n_idx, u in enumerate(self.roots[ell])
            }
        if self.attrs['boundary_condition'] == 'neumann':
            self._normalisations = {
                (ell, n_idx + 1): 2.
                    / self.attrs['boundary_radius'] ** 3
                    / spherical_besselj(ell, u) ** 2
                    / (1 - ell * (ell + 1) / u**2)
                for ell in self.degrees
                for n_idx, u in enumerate(self.roots[ell])
            }

        if self.comm is None or self.comm.rank == 0:
            self.logger.debug("Spectral normalisations computed.")

        return self._normalisations

    def _discretise(self, radius, condition, kmin, kmax, ellmin, ellmax):

        # Avoid duplicate logging entries.
        log = self.comm is None or self.comm.rank == 0

        # Initiate containers.
        self.degrees, self.depths, self.mode_counts = [], [], []
        self.roots = {}

        current_ell = ellmin
        while True:
            # Stop if maximum degree specified is reached.
            if ellmax is not None and current_ell > ellmax:
                if log:
                    self.logger.debug("Specfied maximum degree reached.")
                break

            # Initiate the container and counter for the roots and depth of
            # the current degree.
            u_ell, n_ell = [], 0

            # Iterate through spherical Bessel roots.
            current_root = spherical_besselj_root(
                current_ell, n_ell + 1, derivative=(condition == 'neumann')
            )
            while kmin * radius <= current_root <= kmax * radius:
                u_ell.append(current_root)
                n_ell += 1
                current_root = spherical_besselj_root(
                    current_ell, n_ell + 1, derivative=(condition == 'neumann')
                )

            # If iteration did not proceed, end the entire process.
            if n_ell == 0:
                if log:
                    self.logger.debug(
                        "The last degree is %d.", current_ell - 1
                    )
                break

            self.degrees.append(current_ell)
            self.depths.append(n_ell)
            self.mode_counts.append((2 * current_ell + 1) * n_ell)
            self.roots[current_ell] = np.asarray(u_ell)

            if log:
                self.logger.debug("Degree %d included.", current_ell)

            current_ell += 1

        if log:
            self.logger.info(
                "%s computed: %d degrees and %d modes in total.",
                self, len(self.degrees), sum(self.mode_counts)
            )

    @staticmethod
    def _alias(condition):

        if condition.lower().startswith('d'):
            return 'dirichlet'

        if condition.lower().startswith('n'):
            return 'neumann'

        raise ValueError(f"Unknown boundary `condition`: {condition}.")



import numpy as np
import pytest

from harmonia.algorithms.bases import (
    spherical_besselj,
    spherical_besselj_root,
    spherical_harmonic
)

from . import display_mathematica_query as show_query

TEST_PARAMS = dict(
    ell=5,
    x=10.,
    nmax=3,
    m=2,
    theta=np.pi/3,
    phi=np.pi/5,
)


@pytest.mark.parametrize(
    "ell,x,deriv,value",
    [
        (15, 10., False, 0.00106354),
        (30, 20., True, 0.0000240929),
    ]
)
def test_spherical_besselj(ell, x, deriv, value):

    if deriv:
        show_query(
            f"ReplaceAll[D[SphericalBesselJ[{ell}, z], z], {{z -> {x}}}]"
        )
    else:
        show_query(f"SphericalBesselJ[{ell}, {x}]")

    assert np.isclose(spherical_besselj(ell, x, derivative=deriv), value), \
        "Incorrect evaluaion of spherical Bessel functions."


@pytest.mark.slow
@pytest.mark.parametrize(
    "ell,nmax,deriv,values",
    [
        (5, 4, False, [9.35581, 12.9665, 16.3547, 19.6532]),
        (15, 2, True, [17.4080, 22.6625]),
    ]
)
def test_spherical_besselj_root(ell, nmax, deriv, values):

    if deriv:
        show_query(f"D[SphericalBesselJ[{ell}, x], x] == 0")
    else:
        show_query(f"SphericalBesselJ[{ell}, x] == 0")

    assert np.allclose(
        spherical_besselj_root(ell, nmax, derivative=deriv, only=False),
        values
    ), "Root finding of spherical Bessel functions failed."


@pytest.mark.parametrize(
    "ell,m,theta,phi,value",
    [
        (5, 4, np.pi/3, np.pi/3, -0.206397 - 0.357491*1j),
        (15, -2, np.pi/5, np.pi/6, 0.200298 - 0.346926*1j),
    ]
)
def test_spherical_harmonic(ell, m, theta, phi, value):

    show_query(f"SphericalHarmonicY[{ell}, {m}, {theta}, {phi}]")
    assert np.isclose(spherical_harmonic(ell, m, theta, phi), value), \
        "Incorrect evaluaion of spherical harmonic functions."



"""
Numerical integration (:mod:`~harmonia.algorithms.integration`)
===========================================================================

Integrate numerically in specified coordinate systems.

.. warning::

    Quadrature integration of spherical functions may suffer from poor
    convergence.


**Spherical integrals**

.. autosummary::

    angular_integral
    radial_integral
    pixelated_angular_integral

|

"""
import healpy as hp
import numpy as np
from scipy.integrate import dblquad, quad


def _angular_integrand(phi, theta, func, complex_part):
    r"""Evaluate the angular integrand with the Jacobian factor
    :math:`\sin\theta`.

    Notes
    -----
    Angular arguments are in reverse order for outward double integration.

    Parameters
    ----------
    phi, theta: float, array_like
        Angular coordinates in radians.
    func : callable
        Angular function to be integrated.
    complex_part : {'real', 'imag'}
        Real or imaginary part.

    Returns
    -------
    float, array_like
        Real or imaginary part of the complex angular integrand value.

    """
    if complex_part.lower() == 'real':
        return np.abs(np.sin(theta)) * np.real(func(theta, phi))

    if complex_part.lower() == 'imag':
        return np.abs(np.sin(theta)) * np.imag(func(theta, phi))

    raise ValueError("`complex_part` is neither 'real' nor 'imag'.")


def _radial_integrand(r, func):
    r"""Evaluate the radial integrand with the Jacobian factor :math:`r^2`.

    Parameters
    ----------
    r: float, array_like
        Radial coordinate.
    func : callable
        Radial function to be integrated.

    Returns
    -------
    float, array_like
        Radial integrand value.

    """
    return r**2 * func(r)


def angular_integral(angular_func):
    r"""Compute the full spherical angular integral.

    Notes
    -----
    Arguments :math:`(\theta, \phi)` of `angular_func` must be in radians
    in the domain :math:`[0, \pi] \times [0, 2\pi]`.

    Parameters
    ----------
    angular_func : callable
        Angular function to be integrated.

    Returns
    -------
    complex
        Angular integral value.

    """
    theta_range = (0., np.pi)
    phi_range = (0., 2*np.pi)

    integral_real, _ = dblquad(
        _angular_integrand, *theta_range, *phi_range,
        args=(angular_func, 'real')
    )
    integral_imag, _ = dblquad(
        _angular_integrand, *theta_range, *phi_range,
        args=(angular_func, 'imag')
    )

    return integral_real + 1j*integral_imag


def radial_integral(radial_func, rmax):
    """Compute the radial integral up to the given maximum radius.

    Parameters
    ----------
    radial_func : callable
        Radial function to be integrated.
    rmax : float
        Upper radial limit ``rmax > 0``.

    Returns
    -------
    integral : float
        Radial integral value.

    """
    integral, _ = quad(_radial_integrand, 0., rmax, args=(radial_func,))

    return integral


def pixelated_angular_integral(angular_func, nside):
    r"""Compute the full spherical angular integral with pixelation.

    Notes
    -----
    Arguments :math:`(\theta, \phi)` of `angular_func` must be in radians
    in the domain :math:`[0, \pi] \times [0, 2\pi]`.

    Parameters
    ----------
    angular_func : callable
        Angular function to be integrated.
    nside : int
        'NSIDE' parameter for `healpy` pixelation.

    Returns
    -------
    complex
        Angular integral value.

    """
    num_pixel = hp.nside2npix(nside)

    theta, phi = hp.pix2ang(nside, ipix=range(num_pixel))

    pixel_area = 4 * np.pi / num_pixel

    pixel_value = angular_func(theta, phi)

    return pixel_area * np.sum(pixel_value)



"""
***************************************************************************
Catalogue mapper (:mod:`~harmonia.mapper`)
***************************************************************************

Build and process discrete catalogues into Fourier-space maps.

.. note::

    Unless otherwise specified, the length dimension in the module is
    in units of Mpc/:math:`h`.

"""
from .catalogue_maker import (
    RandomCatalogue,
    SourceCatalogue,
    SphericalFKPCatalogue,
    spherical_indicator,
)
from .map_transform import CartesianMap, SphericalMap



"""Test configuration for :mod:`harmonia`.

"""
import pytest


def pytest_addoption(parser):
    """Add command-line options to `pytest` parser.

    Parameters
    ----------
    parser : :class:`_pytest.config.argparsing.Parser`
        `pytest` parser object.

    """
    parser.addoption(
        '--runslow', action='store_true', default=False, help="Run slow tests."
    )


def pytest_configure(config):
    """Add ini-file options to `pytest` configuration.

    Parameters
    ----------
    config : :class:`_pytest.config.Config`
        `pytest` configuration object.

    """
    config.addinivalue_line('markers', "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    """Modify test collection items.

    Parameters
    ----------
    config : :class:`_pytest.config.Config`
        `pytest` configuration object.
    items : list of :class:`_pytest.nodes.Item`
        `pytest` item objects.

    """
    if config.getoption('--runslow'):
        return

    skip_slow = pytest.mark.skip(
        reason="Use --runslow option to run slow tests."
    )
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)



"""
Definition (:mod:`~harmonia.surveyor.definition`)
===========================================================================

Produce survey definition and specifications for processing and analysing
cosmological data.

.. autosummary::

    generate_mask_by_sky_fraction
    generate_mask_from_map
    generate_selection_by_distribution
    generate_selection_from_samples
    generate_selection_samples

|

"""
# pylint: disable=no-name-in-module
import healpy as hp
import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline as Interpolation
from scipy.special import loggamma
from scipy.stats import gamma, norm

from harmonia.surveyor.coordinates import (
    sky_to_spherical,
    to_box_coords,
    _is_coord_3d,
)


def generate_mask_by_sky_fraction(coord_system, sky_fraction=1.,
                                  box_shift=None):
    """Generate mask function given the sky fraction corresponding to a
    spherical cap (rather than a wedge).

    Parameters
    ----------
    coord_system : {'cartesian', 'spherical', 'sky'}
        Coordinate system of the generated mask function.  If 'sky', the
        returned function can accept either 2-d or 3-d coordinates.
    sky_fraction : float, optional
        Sky coverage, ``0 < sky_fraction <= 1``.
    box_shift : float, array_like or None, optional
        Passed as `box_centre` to :func:`~.coordinates.to_box_coords` so
        that the resulting mask function accepts Cartesian coordinates
        with origin at a box corner.  Ignored unless `coord_system` is
        'cartesian'.

    Returns
    -------
    mask_function : callable
        A mask function accepting `coord_system` as its coordinate system
        and returning boolean values.

    """
    # Any conversion to the 'cartesian' coordinate system as internal
    # `mask_function` is defined for native spherical coordinates.
    if coord_system == 'cartesian':
        apply_to_box_coords_from_native = 'spherical'
    else:
        apply_to_box_coords_from_native = 'null'
        box_shift = None

    # `coords` can be either 'spherical' or 'sky' but the native
    # implementation is 'spherical'; the decorator takes care of
    # 'cartesian' inputs.
    @to_box_coords(apply_to_box_coords_from_native, box_centre=box_shift)
    def mask_function(coords):

        # Transform to spherical surface coords.
        surf_coords = sky_to_spherical(coords) if coord_system == 'sky' \
            else np.atleast_2d(coords)

        return surf_coords[:, -2] <= np.arccos(1 - 2 * sky_fraction)

    return mask_function


def generate_mask_from_map(coord_system, mask_map=None, nside=None,
                           mask_map_file=None, box_shift=None,
                           ret_nside=False):
    """Generate mask function from a veto mask map.

    Parameters
    ----------
    coord_system : {'cartesian', 'spherical', 'sky'}
        Coordinate system of the generated mask function.  If 'sky', the
        returned function can accept either 2-d or 3-d coordinates.
    mask_map : :class:`numpy.ndarray` or None, optional
        A veto array corresponding to pixels in a `healpy` mask map with
        parameter `nside` (default is `None`).  Ignored if
        a `healpy`-generated .fits file for the mask map is provided.
    nside : int or None, optional
        'NSIDE' parameter of the `healpy` mask map (default is `None`).
        Ignored if  a `healpy`-generated .fits file for the
        mask map is provided.
    mask_map_file : *str or* :class:`pathlib.Path`
        `healpy`-generated .fits file for the mask map.  Must contain
        the 'NSIDE' parameter in its header.
    box_shift : float, array_like or None, optional
        Passed as `box_centre` to :func:`~.coordinates.to_box_coords` so
        that the resulting mask function accepts Cartesian coordinates
        with origin at a box corner.  Ignored unless `coord_system` is
        'cartesian'.
    ret_nside : bool, optional
        If `True` (default is `False`), return the `nside` parameter
        from `mask_map` read in.

    Returns
    -------
    mask_function : callable
        A mask function accepting `coord_system` as its coordinate system
        and returning boolean values.

    """
    if mask_map_file is not None:
        mask_map, mask_map_header = hp.read_map(
            mask_map_file, dtype=float, h=True, verbose=False
        )
        nside = dict(mask_map_header)['NSIDE']

    # Any conversion to the 'cartesian' coordinate system as internal
    # `mask_function` is defined for native spherical coordinates.
    if coord_system == 'cartesian':
        apply_to_box_coords_from_native = 'spherical'
    else:
        apply_to_box_coords_from_native = 'null'
        box_shift = None

    # `coords` can be either 'spherical' or 'sky' but the native
    # implementation is 'spherical'; the decorator takes care of
    # 'cartesian' inputs.
    @to_box_coords(apply_to_box_coords_from_native, box_centre=box_shift)
    def mask_function(coords):

        # Transform to spherical surface coords.
        surf_coords = sky_to_spherical(coords) if coord_system == 'sky' \
            else np.atleast_2d(coords)

        pixel = hp.ang2pix(nside, *surf_coords[:, [-2, -1]].T)

        return mask_map[pixel].astype(bool)

    if ret_nside:
        return mask_function, nside

    return mask_function


def generate_selection_by_distribution(coord_scale, distribution, peak,
                                       scale=None, location=None, shape=None):
    """Generate selection function based on a probability distribution
    suitably rescale horizontally and vectically.

    Notes
    -----
    If the generated selection function detects 3-d input coordinate
    arrays, it would assume the coordinates are the Cartesian; otherwise
    it assumes the coordinates are the radial coordinate.

    Parameters
    ----------
    coord_scale : float
        Coordinate scale of the selection function, e.g. the maximum
        comoving radius of a catalogue.
    peak : float
        Peak value of the selection function used for renormalising
        the probability density function by its maximum value.
    distribution : {'gaussian', 'gamma'}
        Distribution to use, either 'gaussian' or 'gamma'.
    scale, location, shape : float or None, optional
        Scale, location and shape parameter of the distribution.  For
        a normal distribution, `scale` is the standard deviation and
        `location` is the mean in units of `coord_scale`; for a gamma
        distribution, see :class:`scipy.stats.gamma`.

    Returns
    -------
    callable
        Normalised selection function of the same coordinate as the sample
        coordinates.

    """
    def rescale_coords(coords):

        coords = np.linalg.norm(coords, axis=-1) if _is_coord_3d(coords) \
                else np.squeeze(coords)

        return coords / coord_scale

    def gaussian_selection(coords):

        r = rescale_coords(coords)

        pdf_peak = 1 / (np.sqrt(2 * np.pi) * scale)

        return peak / pdf_peak * norm.pdf(r, loc=location, scale=scale)

    def gamma_selection(coords):

        r = rescale_coords(coords) / (1 + 5 / np.sqrt(shape))

        pdf_peak = np.exp(
            (shape - 1) * np.log(shape - 1)
            - (shape - 1) - np.log(scale) - loggamma(shape)
        )

        return peak / pdf_peak * gamma.pdf(r, a=shape, scale=scale)

    if distribution == 'gaussian':
        return gaussian_selection

    if distribution == 'gamma':
        return gamma_selection

    raise ValueError(f"Unsupproted distribution: {distribution}.")


def generate_selection_by_cut(low_end, high_end):
    """Generate selection function at a constant value by a cut.

    Notes
    -----
    If the generated selection function detects 3-d input coordinate
    arrays, it would assume the coordinates are the Cartesian; otherwise
    it assumes the coordinates are the radial coordinate.

    Parameters
    ----------
    low_end, high_end : float
        Low and high end of the cut in some selection coordinate.

    Returns
    -------
    callable
        Constant cut selection function.

    """
    def selection_function(coords):

        coord = np.linalg.norm(coords, axis=-1) if _is_coord_3d(coords) \
            else np.squeeze(coords)

        return np.logical_and(
            np.greater_equal(coord, low_end), np.less_equal(coord, high_end)
        )

    return selection_function


def generate_selection_samples(selection_coordinate, coord_scale,
                               redshift_samples, cosmo, sky_fraction=1.,
                               bins=None):
    """Generate selection function samples from redshift samples.

    The selection function is normalised by the mean number density
    inferred from the samples and the sky fraction.

    Parameters
    ----------
    selection_coordinate : {'z', 'r'}
        Coordinate of the generated selection function, either redshift
        'z' or comoving distance 'r'.
    coord_scale : float
        Coordinate scale used to rescale the selection coordinates, e.g.
        the maximum comoving radius of the catalogue to which the
        selection function is applied.
    redshift_samples : float, array_like
        Redshift samples.
    cosmo : :class:`nbodykit.cosmology.cosmology.Cosmology`
        Cosmological model providing the redshift-to-distance conversion
        and thus normalisation of the selection function by volume.
    sky_fraction : float, optional
        The sky coverage of redshift samples as a fraction used to
        normalise the selection function (default is 1.),
        ``0 < sky_fraction <= 1``.
    bins : int, str or None, optional
        Number of bins or binning scheme used to generate the selection
        function samples (see :func:`numpy.histogram_bin_edges`).
        If `None` (default), Scott's rule for binning is used.

    Returns
    -------
    samples, coords : :class:`numpy.ndarray`
        Normalised (dimensionless) selection function samples and
        corresponding coordinates.

    """
    if not 0 < sky_fraction <= 1:
        raise ValueError("Sky fraction must be between 0 and 1.")

    # Perform binning with specified or optimal Scott's binning.
    redshift_samples = redshift_samples[redshift_samples >= 0.]

    bins = bins or 'scott'

    bin_edges = np.histogram_bin_edges(redshift_samples, bins=bins)

    bin_counts, z_edges = np.histogram(redshift_samples, bins=bin_edges)

    z_centres = (z_edges[:-1] + z_edges[1:]) / 2.

    # Find mean number density in each redshift shell bin and normalise
    # by overall density.
    r_edges = cosmo.comoving_distance(z_edges)

    bin_volumes = sky_fraction * (4./3. * np.pi) \
        * (r_edges[1:] ** 3 - r_edges[:-1] ** 3)

    overall_nbar = np.sum(bin_counts) / np.sum(bin_volumes)

    bin_selections = (bin_counts / bin_volumes) / overall_nbar

    bin_centres = z_centres if selection_coordinate == 'z' \
        else cosmo.comoving_distance(z_centres)

    # Generate a selection function of the appropriate coordinate.
    interpolated_selection = Interpolation(bin_centres, bin_selections, ext=1)

    # Attempt resampling and reject volatile ends with negative
    # interpolated values.
    interpolated_selection_samples = interpolated_selection(
        np.linspace(bin_centres.min(), bin_centres.max(), num=len(bin_centres))
    )

    max_point = np.argmax(bin_counts)

    low_partition = interpolated_selection_samples[:max_point]
    high_partition = interpolated_selection_samples[max_point:]

    low_end = np.argmax(low_partition[::-1] < 0)
    high_end = np.argmax(high_partition < 0)

    burnt_remains = slice(max_point - low_end, max_point + high_end)

    coords = bin_centres[burnt_remains] * coord_scale / np.max(bin_centres)
    samples = bin_selections[burnt_remains]

    return samples, coords


def generate_selection_from_samples(sample_selections, sample_coords):
    """Generate selection function interpolated from normalised selection
    function samples at sample coodinates.

    Notes
    -----
    If the generated selection function detects 3-d input coordinate
    arrays, it would assume the coordinates are the Cartesian; otherwise
    it assumes the coordinates are the radial coordinate.

    Parameters
    ----------
    sample_selections, sample_coords : float :class:`numpy.ndarray`
        Selection function samples and coordinates.

    Returns
    -------
    callable
        Normalised selection function of the same coordinate as the sample
        coordinates.

    """
    def selection_function(coords):

        r = np.linalg.norm(coords, axis=-1) if _is_coord_3d(coords) \
            else np.squeeze(coords)

        return Interpolation(sample_coords, sample_selections, ext=1)(r)

    return selection_function



"""
Cosmological likelihoods (:mod:`~harmonia.reader.likelihoods`)
===========================================================================

Construct likelihoods from Spherical Fourier coefficients and Cartesian
power spectrum multipoles of random fields for cosmological parameter
inference.


Probability distributions
---------------------------------------------------------------------------

.. autosummary::

    complex_normal_pdf
    multivariate_normal_pdf
    modified_student_pdf


Likelihood inference
---------------------------------------------------------------------------

.. autosummary::

    spherical_covariance
    cartesian_moments
    LogLikelihood

|

.. |correlator_matrix| replace::

    :meth:`~harmonia.reader.models.SphericalCorrelator.correlator_matrix`

.. |convolved_power_multipoles| replace::

    :meth:`~.reader.models.CartesianMultipoles.convolved_power_multipoles`

"""
import logging
import warnings

# pylint: disable=no-name-in-module
import numpy as np
from scipy.special import loggamma

from harmonia.utils import (
    PositiveDefinitenessWarning,
    is_positive_definite,
    mat_logdet,
)


# Probability distributions
# -----------------------------------------------------------------------------

def _are_valid_moments(first_moment, second_moment):

    # Check dimensions of the expectation and variance are consistent.
    criterion1 = (np.squeeze(first_moment).ndim == 1)
    criterion2 = (np.squeeze(second_moment).ndim == 2)
    criterion3 = (np.shape(first_moment) * 2 == np.shape(second_moment))

    return all([criterion1, criterion2, criterion3])


def chi_square(data_vector, covariance_matrix):
    """Calculate chi-square from zero-centred data vector and its
    covariance matrix.

    Parameters
    ----------
    data_vector : complex, array_like
        Zero-centred data vector.
    covariance_matrix : complex, array_like
        Covariance matrix.

    Returns
    -------
    chi_sq : float :class:`numpy.ndarray`
        Chi-square value.

    """
    data_vector = np.squeeze(data_vector)
    covariance_matrix = np.squeeze(covariance_matrix)
    if not _are_valid_moments(data_vector, covariance_matrix):
        raise ValueError("Check input dimensions.")

    # pylint: disable=unexpected-keyword-arg
    chi_sq = np.dot(
        np.conj(data_vector), np.linalg.solve(covariance_matrix, data_vector)
    )

    return chi_sq


def complex_normal_pdf(data_vector, covariance_matrix, ret_log=True,
                       downscale=None):
    """Compute the complex normal probability density function or its
    natural logarithm given the zero-centred data vector and its
    covariance matrix.

    Parameters
    ----------
    data_vector : complex, array_like
        Zero-centred data vector.
    covariance_matrix : complex, array_like
        Covariance matrix.
    ret_log : bool, optional
        If `True` (default), return logarithmic probability density.
    downscale : float or None, optional
        If not `None` (default), the data vector and covariance matrix are
        simultaneous downscaled to avoid numerical issue.

    Returns
    -------
    float, array_like
        (Logarithmic) probability density.

    """
    data_vector = np.squeeze(data_vector)
    covariance_matrix = np.squeeze(covariance_matrix)
    if not _are_valid_moments(data_vector, covariance_matrix):
        raise ValueError("Check input dimensions.")

    dim = np.size(data_vector)

    log_normalisation = dim * np.log(np.pi)
    if downscale is not None:
        data_vector = data_vector / downscale
        covariance_matrix = covariance_matrix / downscale ** 2
        log_normalisation -= 2 * dim * np.log(downscale)

    if not is_positive_definite(covariance_matrix):
        warnings.warn(
            "`covariance_matrix` is not positive definite.",
            PositiveDefinitenessWarning
        )

    density = \
        - log_normalisation \
        - mat_logdet(covariance_matrix) \
        - np.real_if_close(
            chi_square(data_vector, covariance_matrix), tol=10**10
        )

    return density if ret_log else np.exp(density)


def multivariate_normal_pdf(data_vector, expectation_vector, covariance_matrix,
                            ret_log=True):
    """Compute the multivariate normal probability density function or its
    natural logarithm given the data vector, its mean vector and covariance
    matrix.

    Parameters
    ----------
    data_vector : float, array_like
        Data vector.
    expectation_vector : float, array_like
        Mean vector.
    covariance_matrix : float, array_like
        Covariance matrix.
    ret_log : bool, optional
        If `True` (default), return logarithmic probability density.

    Returns
    -------
    density : float, array_like
        Logarithmic probability density value.

    """
    data_vector = np.squeeze(data_vector)
    expectation_vector = np.squeeze(expectation_vector)
    covariance_matrix = np.squeeze(covariance_matrix)
    if not _are_valid_moments(data_vector, covariance_matrix) or \
            not _are_valid_moments(expectation_vector, covariance_matrix):
        raise ValueError("Check input dimensions.")

    dim = np.size(data_vector)

    log_normalisation = dim * np.log(2*np.pi)

    log_determinant = mat_logdet(covariance_matrix)

    chi_sq = chi_square(data_vector - expectation_vector, covariance_matrix)

    density = - (log_normalisation + log_determinant + chi_sq) / 2

    return density if ret_log else np.exp(density)


def modified_student_pdf(data_vector, expectation_vector, covariance_matrix,
                         degree, ret_log=True):
    """Compute the multivariate modified Student probability density
    function or its natural logarithm given the data vector, its mean
    vector and covariance matrix.

    Parameters
    ----------
    data_vector : float, array_like
        Data vector.
    expectation_vector : float, array_like
        Mean vector.
    covariance_matrix : float, array_like
        Covariance matrix.
    degree : int
        The degree number.  This could be the number of empirical
        covariance matrices used to obtain the estimated
        `covariance_matrix`.
    ret_log : bool, optional
        If `True` (default), return logarithmic probability density.

    Returns
    -------
    float, array_like
        (Logarithmic) probability density value.

    """
    data_vector = np.squeeze(data_vector)
    expectation_vector = np.squeeze(expectation_vector)
    covariance_matrix = np.squeeze(covariance_matrix)
    if not _are_valid_moments(data_vector, covariance_matrix) \
            or not _are_valid_moments(expectation_vector, covariance_matrix):
        raise ValueError("Check input dimensions.")

    dim = np.size(data_vector)

    log_normalisation = dim / 2. * np.log((degree - 1) * np.pi) \
        + loggamma((degree - dim) / 2.) \
        - loggamma(degree / 2.)

    log_determinant = mat_logdet(covariance_matrix)

    log_pseudo_chisq = degree / 2. * np.log(
        1 + chi_square(
            data_vector - expectation_vector, covariance_matrix
        ) / (degree - 1)
    )

    density = - (log_normalisation + log_determinant / 2 + log_pseudo_chisq)

    return density if ret_log else np.exp(density)


# Likelihoods
# -----------------------------------------------------------------------------

class LikelihoodWarning(UserWarning):
    """Likelihood evaluation warning.

    """


def spherical_covariance(pivot, spherical_model, **kwargs):
    r"""Compute the parametrised covariance matrix of spherical Fourier
    coefficients.

    Parameters
    ----------
    pivot : {'natural', 'spectral'}
        Pivot order for vectorisation.
    spherical_model : :class:`~harmonia.reader.models.SphericalCorrelator`
        Spherical correlator base model.
    **kwargs
        Parameters (other than `pivot`) to be passed to |correlator_matrix|
        of `spherical_correlator`.

    Returns
    -------
    covariance_matrix : complex :class:`numpy.ndarray`
        Covariance matrix.

    See Also
    --------
    :class:`~harmonia.reader.models.SphericalCorrelator`

    """
    covariance_matrix = spherical_model.correlator_matrix(pivot, **kwargs)

    return covariance_matrix


def cartesian_moments(pivot, orders, cartesian_model,
                      covariance_estimator=None, mode_counts=None, **kwargs):
    """Compute the parametrised mean and covariance of Cartesian
    power spectrum multipoles.

    Parameters
    ----------
    pivot : {'order', 'wavenumber'}
        Pivot order for vectorisation.
    orders : list of int
        Orders of the power spectrum multipoles.
    cartesian_model : :class:`~.models.CartesianMultipoles`
        Cartesian power multipoles base model.
    covariance_estimator : :class:`~.CovarianceEstimator` *or None, optional*
        Cartesian power multipole covariance estimator.  Its
        :attr:`wavenumbers` must match wavenumbers associated
        with `cartesian_model`.  If `None`, no correlation between power
        spectrum multipoles is assumed but `mode_counts` must be provided
        for calculating the power spectrum variance.
    mode_counts : int, array_like or None, optional
        Number of independent modes for each power spectrum measurement
        (default is `None`) used to calculate the power spectrum variance.
        Ignored if `covariance_estimator` is provided.
    **kwargs
        Parameters (other than `orders`) to be passed to
        |convolved_power_multipoles| of `cartesian_model`.

    Returns
    -------
    expectation : float :class:`numpy.ndarray`
        Power spectrum expectation at the specified wavenumbers.
    covariance : float :class:`numpy.ndarray`
        Power spectrum variance at the specified wavenumbers.

    """
    expectation = cartesian_model.convolved_power_multipoles(
        orders, **kwargs
    ).vectorise(pivot)

    # Check model and estimator wavenumbers agree.
    if covariance_estimator is None:
        covariance = expectation ** 2 / np.asarray(mode_counts)
    else:
        assert np.allclose(
            cartesian_model.attrs['wavenumbers'],
            covariance_estimator.wavenumbers,
            atol=0.001
        ), (
            "The wavenumbers at which the Cartesian power multipole model "
            "is evaluated must match the wavenumbers at which "
            "the fiducial covariance matrix is estimated."
        )

        fiducial_expectation = \
            covariance_estimator.get_fiducial_vector(pivot)
        fiducial_covariance = \
            covariance_estimator.get_fiducial_covariance(pivot)

        covariance = np.linalg.multi_dot([
            np.diag(expectation / fiducial_expectation),
            fiducial_covariance,
            np.diag(expectation / fiducial_expectation)
        ])

    return expectation, covariance


class LogLikelihood:
    """Construct the logarithmic likelihood function from
    cosmological data.

    Parameters
    ----------
    spherical_data : :class:`~.arrays.SphericalArray` *or None, optional*
        Spherical Fourier coefficient data (default is `None`).
    cartesian_data : :class:`~.arrays.CartesianArray` *or None, optional*
        Spherical Fourier coefficient data (default is `None`).
    covariance_estimator : :class:`~.CovarianceEstimator` *or None, optional*
        Cartesian multipole covariance estimator (default is `None`).
    mode_counts : int, array_like or None, optional
        Number of independent modes for each Cartesian data point (default
        is `None`) as an alternative to `covariance_estimator`.  Ignored
        if `covariance_estimator` is provided.
    base_spherical_model : :class:`~.SphericalCorrelator` *or None, optional*
        Baseline spherical correlator model (default is `None`).
    base_cartesian_model : :class:`~.CartesianMultipoles` *or None, optional*
        Baseline Cartesian multipole model (default is `None`).
    spherical_pivot : {'natural', 'spectral'}, optional
        Pivot order for spherical map data vectorisation (default is
        'natural').
    cartesian_pivot : {'order', 'wavenumber'}, optional
        Pivot order for Cartesian map data vectorisation (default is
        'order').
    nbar : float or None, optional
        Mean particle number density (in cubic :math:`h`/Mpc).  If
        `None` (default), shot noise is neglected.
    contrast : float or None, optional
        If not `None` (default), this adds additional shot noise level
        ``1 / (contrast * nbar)`` due to a FKP-style random catalogue.
    tracer_p : float, optional
        Tracer-dependent parameter for bias modulation by `f_nl`
        (default is 1.).
    comm : :class:`mpi4py.MPI.Comm` *or None, optional*
        MPI communicator (default is `None`).

    Attributes
    ----------
    attrs : dict
        Directory holding input parameters not corresponding to any of
        the following attributes.
    spherical_data : :class:`~.algorithms.arrays.SphericalArray` or None
        Spherical Fourier coefficient data.
    cartesian_data : :class:`~.algorithms.arrays.CartesianArray` or None
        Spherical Fourier coefficient data.
    covariance_estimator : :class:`~.CovarianceEstimator` or None
        Cartesian multipole covariance estimator.
    base_spherical_model : :class:`~.SphericalCorrelator` or None
        Baseline spherical correlator model.
    base_cartesian_model : :class:`~.CartesianMultipoles` or None
        Baseline Cartesian multipole model.

    """

    def __init__(self, spherical_data=None, cartesian_data=None,
                 covariance_estimator=None, mode_counts=None,
                 base_spherical_model=None, base_cartesian_model=None,
                 spherical_pivot='natural', cartesian_pivot='order',
                 nbar=None, contrast=None, tracer_p=1., comm=None):

        self.logger = logging.getLogger(self.__class__.__name__)
        self.comm = comm

        self.attrs = {
            'spherical_pivot': spherical_pivot,
            'cartesian_pivot': cartesian_pivot,
            'mode_counts': mode_counts,
            'nbar': nbar,
            'contrast': contrast,
            'tracer_p': tracer_p,
        }

        self.spherical_data = spherical_data
        self.cartesian_data = cartesian_data
        self.covariance_estimator = covariance_estimator
        self.base_spherical_model = base_spherical_model
        self.base_cartesian_model = base_cartesian_model

    def spherical_map_likelihood(self, b_1, f_nl, exclude_degrees=(),
                                 compression_matrix=None, **kwargs):
        """Evaluate the spherical map logarithmic likelihood.

        Parameters
        ----------
        b_1 : float
            Scale-independent linear bias.
        f_nl : float or None
            Local primordial non-Gaussianity.
        exclude_degrees : tuple of int, optional
            If not empty (default), modes whose spherical degree
            match one of its elements are removed from the likelihood.
        compression_matrix : :class:`numpy.ndarray` *or None*, optional
            If not `None` (default), both the data vector and the model
            covariance matrix are processed for data compression.  This
            must be compatible with `exclude_degrees`, i.e. it accounts
            for elements removed from the data vector and covariance
            matrix by `exclude_degrees`.
        **kwargs
            Additional parameters to be passed to |correlator_matrix| of
            :attr:`base_spherical_model`.

        Returns
        -------
        log_likelihood : float
            Logarithmic likelihood.

        See Also
        --------
        :class:`~harmonia.surveyor.synthesis.generate_compression_matrix`

        :class:`~harmonia.reader.likelihoods.spherical_covariance`

        |correlator_matrix|

        """
        _OVERFLOW_DOWNSCALE = 10**4

        data_vector = \
            self.spherical_data.vectorise(self.attrs['spherical_pivot'])

        covariance_matrix = spherical_covariance(
            self.attrs['spherical_pivot'], self.base_spherical_model,
            b_1=b_1, f_nl=f_nl,
            nbar=self.attrs['nbar'],
            contrast=self.attrs['contrast'],
            tracer_p=self.attrs['tracer_p'],
            **kwargs
        )

        # pylint: disable=no-member
        if exclude_degrees:
            deselector = np.logical_and.reduce([
                self.spherical_data.array['index'][:, 0] == deg
                for deg in exclude_degrees
            ])

            data_vector = data_vector[~deselector]
            covariance_matrix = \
                covariance_matrix[~deselector, :][:, ~deselector]

        if compression_matrix is not None:
            data_vector = np.linalg.multi_dot([
                compression_matrix, data_vector
            ])
            covariance_matrix = np.linalg.multi_dot([
                compression_matrix,
                covariance_matrix,
                np.conj(compression_matrix.T)
            ])

        log_likelihood = complex_normal_pdf(
            data_vector, covariance_matrix, downscale=_OVERFLOW_DOWNSCALE,
        )

        return log_likelihood

    def cartesian_map_likelihood(self, b_1, f_nl, orders=None,
                                 num_samples=None, **kwargs):
        """Evaluate the Cartesian map logarithmic likelihood.

        Parameters
        ----------
        b_1 : float
            Scale-independent linear bias of the tracer particles.
        f_nl : float or None
            Local primordial non-Gaussianity.
        orders : list of int or None, optional
            Orders of the power spectrum multipoles.  If `None` (default),
            only the monopole is used.
        num_samples : int or None, optional
            If `None` (default), the normal distribution is used without
            correction for covariance estimation uncertainty; otherwise
            it is passed as `degree` to :func:`modified_student_pdf` for
            covariance estimation uncertainty correction [1]_.
        **kwargs
            Additional parameters to be passed to
            |convolved_power_multipoles| of :attr:`base_cartesian_model`.

        Returns
        -------
        log_likelihood : float
            Logarithmic likelihood.

        See Also
        --------
        :class:`~harmonia.reader.likelihoods.cartesian_moments`

        |convolved_power_multipoles|


        .. [1] Sellentin E. & Heavens A. F., 2016. MNRAS 456(1), L132–L136.
           [arXiv: `1511.05969 <https://arxiv.org/abs/1511.05969>`_]

        """
        orders = orders or [0]

        data_vector = \
            self.cartesian_data.vectorise(self.attrs['cartesian_pivot'])

        expectation_vector, covariance_matrix = cartesian_moments(
            self.attrs['cartesian_pivot'], orders, self.base_cartesian_model,
            covariance_estimator=self.covariance_estimator,
            mode_counts=self.attrs['mode_counts'],
            b_1=b_1, f_nl=f_nl,
            nbar=self.attrs['nbar'],
            contrast=self.attrs['contrast'],
            tracer_p=self.attrs['tracer_p'],
            **kwargs
        )

        if self.covariance_estimator is not None and num_samples is not None:
            log_likelihood = modified_student_pdf(
                data_vector, expectation_vector, covariance_matrix,
                degree=num_samples
            )
        else:
            log_likelihood = multivariate_normal_pdf(
                data_vector, expectation_vector, covariance_matrix
            )

        return log_likelihood



"""
Utilities (:mod:`~harmonia.utils`)
===========================================================================

Provide utilities for processing and logging as well as common mathematical
function and algorithms.


**Processing and monitoring**

.. autosummary::

    Progress
    setup_logger
    clean_warning_format
    restore_warnings
    mpi_compute


**Mathematics**

.. autosummary::

    const_function
    normalise_vector
    binary_search
    covar_to_corr
    mat_logdet
    PositiveDefinitenessWarning
    is_positive_definite

|

"""
import logging
import random
import sys
import time
import warnings

import numpy as np
from tqdm import tqdm

__all__ = [
    'Progress',
    'setup_logger',
    'clean_warning_format',
    'restore_warnings',
    'mpi_compute',
    'const_function',
    'binary_search',
    'covar_to_corr',
    'mat_logdet',
    'PositiveDefinitenessWarning',
    'is_positive_definite',
]


# Processing and monitoring utilities
# -----------------------------------------------------------------------------

class Progress:
    """Progress status of tasks.

    This is an alternative to ``tqdm`` for cases where progress is not
    uniform and only needs to be infrequently reported to a logger.  If
    multiple parallel processes exist, progress status is only reported
    for the first and last of them.

    Parameters
    ----------
    task_length : int
        Total number of tasks.
    num_checkpts : int, optional
        Number of checkpoints for reporting progress (default is 4).
    process_name : str or None, optional
        If not `None` (default), this is the process name to be logged.
    logger : :class:`logging.Logger` *or None, optional*
        Logger.  If `None` (default), a print statement is issued.
    comm : :class:`mpi4py.MPI.Comm` *or None, optional*
        MPI communicator (default is `None`).
    root : int, optional
        Root process number (default is 0).

    Attributes
    ----------
    process_name : str or None, optional
        If not `None` (default), this is the process name to be logged.
    task_length : int
        Total number of tasks.
    progress_checkpts : float
        Scheduled progress check points, ``0 < progress_checkpts <= 1``.
    last_checkpt : int
        Index of the last passed checkpoint,
        ``0 <= last_checkpt <= num_checkpts``.

    Examples
    --------
    >>> ntasks = 100
    >>> p = Progress(ntasks, process_name='null test')
    >>> for task_idx in range(ntasks):
    ...     p.report(task_idx)
    Progress for the single 'null test' process: 25% computed.
    Progress for the single 'null test' process: 50% computed.
    Progress for the single 'null test' process: 75% computed.
    Progress for the single 'null test' process: 100% computed.

    """

    def __init__(self, task_length, num_checkpts=4, process_name=None,
                 logger=None, comm=None, root=0):

        self.process_name = process_name
        self.task_length = task_length
        self.logger = logger

        if self.process_name is None:
            self._proc_name = ""
        else:
            self._proc_name = "'{}' ".format(process_name)

        if comm is None:
            self._which_proc = 'single'
        else:
            if comm.rank == root:
                tracked_rank = random.randint(0, comm.size - 1)
            else:
                tracked_rank = None
            tracked_rank = comm.bcast(tracked_rank, root=root)

            self._which_proc = _num_to_ordinal(tracked_rank + 1) \
                if comm.rank == tracked_rank else None

        self.progress_checkpts = \
            np.linspace(1. / num_checkpts, 1., num=num_checkpts)
        self.last_checkpt = 0

        self._progressor = self._initialise()

    def report(self, current_position):
        """Report the current position in the tasks.

        Parameters
        ----------
        current_position : int
            Index of the current position in the tasks (starting from 0).

        """
        next(self._progressor)
        self._progressor.send(current_position)

    def _initialise(self):

        while True:
            current_idx = yield

            current_progress = (current_idx + 1) / self.task_length
            place_in_checkpts = np.searchsorted(
                self.progress_checkpts, current_progress, side='right'
            )

            if place_in_checkpts > self.last_checkpt \
                    and self._which_proc is not None:
                if self.logger is None:
                    print(
                        "Progress for the {} {}process: {:.0f}% computed."
                        .format(
                            self._which_proc, self._proc_name,
                            100 * current_progress
                        )
                    )
                else:
                    self.logger.info(
                        "Progress for the %s %sprocess: %.0f%% computed.",
                        self._which_proc, self._proc_name,
                        100 * current_progress
                    )
                self.last_checkpt = place_in_checkpts
            yield


class _LoggerFormatter(logging.Formatter):
    """Customised logging formatter.

    """

    _start_time = time.time()

    def format(self, record):
        """Modify the default logging record by adding elapsed time in
        hours, minutes and seconds.

        Parameters
        ----------
        record : :class:`Logging.LogRecord`
            Default logging record object.

        Returns
        -------
        str
            Modified record message with elapsed time.

        """
        elapsed_time = record.created - self._start_time
        h, remainder_time = divmod(elapsed_time, 3600)
        m, s = divmod(remainder_time, 60)

        record.elapsed = "(+{}:{:02d}:{:02d})".format(int(h), int(m), int(s))

        return logging.Formatter.format(self, record)


def setup_logger():
    """Return the root logger formatted with elapsed time and piped
    to ``stdout``.

    Returns
    -------
    logger : :class:`logging.Logger`
        Formatted root logger.

    """
    logger = logging.getLogger()
    logging_handler = logging.StreamHandler(sys.stdout)
    logging_formatter = _LoggerFormatter(
        fmt='[%(asctime)s %(elapsed)s %(levelname)s] %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )

    logging_handler.setFormatter(logging_formatter)
    logger.addHandler(logging_handler)

    return logger


# pylint: disable=unused-argument
def clean_warning_format(message, category, filename, lineno, line=None):
    """Clean warning message format.

    Parameters
    ----------
    message, category, filename, lineno : str
        Warning message, warning catagory, origin filename, line number.
    line : str or None, optional
        Source code line to be included in the warning message (default is
        `None`).

    Returns
    -------
    str
        Warning message format.

    """
    filename = filename if "harmonia" not in filename \
        else "".join(filename.partition("harmonia")[1:])

    return '%s:%s: %s: %s\n' % (filename, lineno, category.__name__, message)


def restore_warnings(captured_warnings):
    """Emit captured warnings.

    Parameters
    ----------
    captured_warnings : *list of* :class:`warnings.WarningMessage`
        List of recorded warnings as returned by
        ``warnings.catch_warnings(record=True)``.

    """
    for record in captured_warnings:
        warnings.showwarning(
            record.message, record.category, record.filename, record.lineno,
            file=record.file, line=record.line
        )


def _allocate_tasks(total_task, total_proc):
    """Allocate tasks to processes for parallel computation.

    If `total_proc` processes share `total_task` tasks, then this decides
    the numbers of tasks, `tasks`, different processes receive: the
    rank-``i`` process receives ``tasks[i]`` many tasks.

    Parameters
    ----------
    total_task : int
        Total number of tasks.
    total_proc : int
        Total number of processes.

    Returns
    -------
    tasks : list of int
        Number of tasks for each process.

    """
    try:
        total_task, total_proc = int(total_task), int(total_proc)
    except TypeError as error:
        raise TypeError(
            "`total_task` and `total_proc` must have integer values."
        ) from error

    num_task_remaining, num_proc_remaining, tasks = total_task, total_proc, []

    while num_task_remaining > 0:
        num_task_assigned = num_task_remaining // num_proc_remaining
        tasks.append(num_task_assigned)
        num_task_remaining -= num_task_assigned
        num_proc_remaining -= 1

    return tasks


def _allocate_segments(tasks=None, total_task=None, total_proc=None):
    """Allocate segments of tasks to each process by the number of tasks it
    receives and its rank.

    For instance, if the rank-``i`` process receives ``tasks[i]`` tasks
    (e.g. assigned by :func:`_allocate_tasks`), then this function assigns
    a slice of the indexed tasks it should receive, with the indices in
    ascending order in correspondence with ranks of the processes.

    Parameters
    ----------
    tasks : list of int or None, optional
        Number of tasks each process receives.  Cannot be `None` if either
        `total_task` or `total_proc` is `None`.  If not `None`,
        `total_task` and `total_proc` are both ignored.
    total_task : int or None, optional
        Total number of tasks.  Ignored if `tasks` is not `None`, otherwise
        cannot be `None`.
    total_proc : int or None, optional
        Total number of processes.  Ignored if `tasks` is not `None`,
        otherwise cannot be `None`.

    Returns
    -------
    segments : list of slice
        Index slice of the segment of tasks that each process should
        receive.

    """
    if tasks is None:
        tasks = _allocate_tasks(total_task, total_proc)
    if total_proc is None:
        total_proc = len(tasks)

    breakpoints = np.insert(np.cumsum(tasks), 0, values=0)
    segments = [
        slice(breakpoints[rank], breakpoints[rank + 1])
        for rank in range(total_proc)
    ]

    return segments


def mpi_compute(data_array, mapping, comm=None, root=0, process_name=None,
                update_rate=None):
    """Multiprocess mapping of data with optional progress bar.

    For each map to be applied, the input data array is scattered over the
    first axis for computation on difference process, and the computed
    results are gathered in the exact structure of the input data array on
    the root process.

    Parameters
    ----------
    data_array : array_like
        Data array.
    mapping : callable
        Mapping to be applied.
    comm : :class:`mpi4py.MPI.Comm` *or None, optional*
        MPI communicator.  If `None`, no multiprocessing is performed.
    root : int, optional
        Rank of the process taken as the root process (default is 0).
    process_name : str or None
        If `None` (default), no progress bar is displayed; else this is
        the process name to be displayed.
    update_rate : int or None, optional
        If not `None` (default), this is the progress bar update rate
        (in inverse seconds).  Has no effect if `process_name` is
        not provided.

    Returns
    -------
    output_array : array_like or None
        Output data processed from `mapping`.  Returns `None` for process
        ranks other than `root`.

    """
    process_name = "" if process_name is None \
        else process_name[0].upper() + process_name[1:]
    update_rate = 1 if update_rate is None else update_rate

    ## Deal with single processes.

    if comm is None or comm.size == 1:
        if process_name:
            output_array = list(tqdm(
                map(mapping, data_array),
                total=len(data_array), mininterval=update_rate,
                desc=process_name, file=sys.stdout
            ))
        else:
            output_array = list(map(mapping, data_array))
        return output_array

    ## Deal with multiple processes.

    if root + 1 > comm.size:
        root = 0
        warnings.warn(
            "Root rank reset to 0 as `root` exceeds the number of processes."
        )

    # Decides which process to track.
    if comm.rank == root:
        tracked_rank = random.randint(0, comm.size - 1)
    else:
        tracked_rank = None

    tracked_rank = comm.bcast(tracked_rank, root=root)
    tracked_process = " ({} process)".format(_num_to_ordinal(tracked_rank + 1))

    segments = _allocate_segments(
        total_task=len(data_array), total_proc=comm.size
    )

    data_chunk = data_array[segments[comm.rank]]

    comm.Barrier()
    if process_name and comm.rank == tracked_rank:
        output = list(tqdm(
            map(mapping, data_chunk),
            total=len(data_chunk), mininterval=update_rate,
            desc=(process_name + tracked_process).strip(), file=sys.stdout
        ))
    else:
        output = list(map(mapping, data_chunk))
    comm.Barrier()

    output = comm.gather(output, root=root)

    output_array = None
    if comm.rank == root:
        try:
            output_array = np.concatenate(output, axis=0)
        except ValueError:
            output_array = [
                array_in_block
                for output_block in output
                for array_in_block in output_block
            ]

    output_array = comm.bcast(output_array, root=root)

    return output_array


def _num_to_ordinal(num):

    _determine_suffix = lambda n: {
        1: "st", 2: "nd", 3: "rd", 11: "th", 12: "th", 13: "th"
    }.get(n % 100 if n % 100 < 20 else n % 10, "th")

    return "{}{}".format(num, _determine_suffix(num))


# Mathematical functions and algorithms
# -----------------------------------------------------------------------------

def const_function(const):
    """Return a constant function with arbitrary arguments.

    Parameters
    ----------
    const : float
        Constant value.

    Returns
    -------
    callable
        Constant function.

    """
    return lambda *args, **kwargs: const


def normalise_vector(vector_array):
    """Normalise vector(s).

    Parameters
    ----------
    vector_array : float, array_like
        An array of row vectors.

    Returns
    -------
    float :class:`numpy.ndarray`
        Normalise vector array.

    """
    norms = np.linalg.norm(vector_array, axis=-1)

    return vector_array / norms[:, None]


def _scan_interval(func, a, b, dx):
    """Scan an interval from the lower end to detect sign changes.

    Parameters
    ----------
    func : callable
        Function whose sign-change interval is to be found.
    a, b: float
        Initial interval end points, ``a < b``.
    dx : float
        Increment from the interval lower end, ``dx > 0``.

    Returns
    -------
    x_low, x_high : float or None
        End points for a sign-change interval, ``x_low < x_high``.
        `None` if the result is null.

    """
    # Incremental interval.
    x_low, x_high = a, a + dx
    f_low, f_high = func(x_low), func(x_high)
    # Continue interval increments unless sign changes.
    while f_low * f_high >= 0:
        # Terminate when incremental interval goes outside the
        # overall interval.
        if x_low >= b:
            return None, None
        x_low, x_high = x_high, x_high + dx
        f_low, f_high = f_high, func(x_high)

    return x_low, x_high


def _find_root(func, x_low, x_high, convergence=1.e-9):
    """Bisection method for root finding.

    Parameters
    ----------
    func : callable
        Function whose zero bracket is to be found.
    x_low, x_high: float
        Initial interval end points.
    convergence : float, optional
        Convergence precision for setting maximum iteration number
        (default is 1.e-9).

    Returns
    -------
    float or None
        Single possible root.

    """
    f_low, f_high = func(x_low), func(x_high)

    # Trivial termination scenarios.
    if f_low == 0:
        return x_low
    if f_high == 0:
        return x_high
    if f_low * f_high > 0:
        warnings.warn("Root is not bracketed.", RuntimeWarning)
        return None

    maxiter = int(np.log((x_high - x_low)/convergence) / np.log(2) + 1)
    for _ in range(maxiter):
        x_middle = (x_low + x_high) / 2
        f_middle = func(x_middle)
        if f_middle == 0:
            return x_middle
        if f_high * f_middle < 0:
            x_low = x_middle
            f_low = f_middle
        else:
            x_high = x_middle
            f_high = f_middle

    return (x_low + x_high) / 2


def binary_search(func, a, b, maxnum=None, precision=1.e-6):
    """Binary seach for all zeros of a function in a real interval.

    Parameters
    ----------
    func : callable
        Function whose zeros are to be found.
    a, b : float
        Interval end points, ``a < b``.
    maxnum : int or None, optional
        Maximum number of zeros needed from below (default is `None`).
    precision : float, optional
        Desired absolute precision of the zeros (default is 1.e-8).

    Returns
    -------
    roots : float :class:`numpy.ndarray` or None
        Possible roots.

    Raises
    ------
    ValueError
        If the initial interval covers only one point (``a == b``).

    """
    # Maximum number of roots for definite termination.
    if maxnum is None:
        maxnum = np.iinfo(np.int32).max

    # Check overall interval is valid.
    if a == b:
        raise ValueError(
            f"Initial interval covers only one point: [{a}, {b}]."
        )
    if a > b:
        a, b = b, a
        warnings.warn(
            f"Initial interval [{a}, {b}] reordered to [{b}, {a}]."
        )

    roots = []
    while len(roots) < maxnum:
        x_low, x_high = _scan_interval(func, a, b, precision)
        if x_low is not None:
            root = _find_root(func, x_low, x_high)
            if root is not None:
                roots.append(round(root, int(- np.log10(precision))))
            a = x_high
        else:
            break

    return np.asarray(roots, dtype=float)


def covar_to_corr(covar):
    """Convert a covariance matrix to its correlation matrix.

    Parameters
    ----------
    covar : complex, array_like
        Covariance matrix.

    Returns
    -------
    corr : :class:`numpy.ndarray`
        Correlation matrix.

    """
    inv_diag = np.diag(np.power(np.abs(np.diag(covar)), - 1./2.))
    corr = inv_diag @ covar @ inv_diag

    return corr


def mat_logdet(matrix):
    """Calculate logarithm of the determinant of a matrix.

    Parameters
    ----------
    matrix : float or complex, array_like
        Matrix.

    Returns
    -------
    log_det : float
        Logarithm of the matrix determinant.

    """
    _, log_det = np.linalg.slogdet(matrix)

    return log_det


class PositiveDefinitenessWarning(UserWarning):
    """Emit a warning when a matrix is not positive definite.

    """


def is_positive_definite(matrix):
    """Check the positive definiteness of a square matrix by attempting a
    Cholesky decomposition.

    Parameters
    ----------
    matrix : float or complex, array_like
        Matrix.

    Returns
    -------
    bool
        Positive definiteness.

    """
    try:
        _ = np.linalg.cholesky(matrix)
    except np.linalg.LinAlgError:
        return False
    else:
        return True



"""
Background geometry (:mod:`~harmonia.cosmology.geometry`)
===========================================================================

Compute geometrical quantities in a cosmological background.

.. autosummary::

    redshift_from_distance
    differential_AP_distortion

|

"""
import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline as Spline


def redshift_from_distance(cosmo, lg_num_sample=5):
    """Invert redshift-to-distance relationship of a cosmological model
    to redshift-from-distance.

    Notes
    -----
    This is useful when the Alcock--Paczynski effect needs to be included
    in modelling.  Only valid for redshift between 1.e-3 and 100.

    Parameters
    ----------
    cosmo : :class:`nbodykit.cosmology.cosmology.Cosmology`
        Cosmological model.
    lg_num_sample : float, optional
        Base-10 logarithm of the number of redshift points to sample
        the comoving distance as a function of redshift (default
        is 5, i.e. 100000 samle points).

    Returns
    -------
    callable
        Redshift-from-distance function.

    """
    Z_LOG_RANGE = (-3, 2)

    z_samples = np.logspace(*Z_LOG_RANGE, num=10**lg_num_sample)
    r_samples = cosmo.comoving_distance(z_samples)

    return Spline(r_samples, z_samples, ext='raise')


# pylint: disable=assignment-from-no-return
def differential_AP_distortion(fiducial_z_to_r, variable_z_to_r,
                               max_redshift=10., lg_num_sample=5):
    """Compute the differential Alcock--Paczynski distortion between a
    fiducial and a cosmological redshift-to-distance conversion as a
    fuction of redshift.

    Parameters
    ----------
    fiducial_z_to_r : callable
        Fiducial redshift-to-distance conversion.
    variable_z_to_r : callable
        Variable redshift-to-distance conversion.
    max_redshift : float, optional
        Maximum redshift to sample for interpolation (default is 10).
    lg_num_sample : float, optional
        Base-10 logarithm of the number of redshift points to sample
        the differential distortion as a function of redshift (default
        is 5, i.e. 100000 samle points).

    Returns
    -------
    callable
        Differential distortion as a fuction of redshift.

    """
    z_samples = np.linspace(0., max_redshift, 10 ** lg_num_sample)

    r_tilde = fiducial_z_to_r(z_samples)
    r = variable_z_to_r(z_samples)

    with np.errstate(divide='ignore'):
        dr_tilde_over_dr = np.gradient(r_tilde) / np.gradient(r)

    # Filter out where possibly the gradient of r(z) is close to zero.
    valid_samples = np.isfinite(dr_tilde_over_dr)

    z_samples = z_samples[valid_samples]
    dr_tilde_over_dr = dr_tilde_over_dr[valid_samples]

    return Spline(z_samples, dr_tilde_over_dr, check_finite=False)



"""
***************************************************************************
Unit testing (:mod:`~harmonia.tests`)
***************************************************************************

Unit testing with `pytest` in Python 3.5 or above.  Any public function
is tested only if: it is not a method of a class; has a yield or return
statement; and do not return a string, a callable or a class instance.
Any class that does not depend on a class from an external package is
tested for their public methods, properties and most of the magic methods.

"""


def display_mathematica_query(message: str):
    """Print strings to be used as Mathematica/WolframAlpha queries.

    """
    print("Query: \n{}\n".format(message))


class NamedFunction:
    """Named functions.

    Parameters
    ----------
    name: Name of the function.
    func: Function.

    Attributes
    ----------
    name: Name of the function.
    func: Function.

    """
    name: str
    func: callable

    def __init__(self, name: str, func: callable):
        self.name = name
        self.func = func

    def __str__(self):
        return self.name

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)



import pytest

from harmonia.mapper.catalogue_maker import spherical_indicator


@pytest.mark.parametrize(
    "vec,r,value",
    [
        ([1, 0, -1], 1, False),
    ]
)
def test_spherical_indicator(vec, r, value):
    assert (spherical_indicator(vec, r) == value), \
        "Incorrect indication for spherical domain."



# pylint: disable=missing-module-docstring
import setuptools

with open("README.md", 'r') as description:
    long_description = description.read()

with open("requirements.txt", 'r') as dependencies:
    requirements = [pkg.strip() for pkg in dependencies]

with open("version.txt", 'r') as version_info:
    version_tag, version = [v.strip() for v in version_info]
    if version_tag == 'latest':
        branch = 'master'
    else:
        branch = version_tag

setuptools.setup(
    name="HarmoniaCosmo",
    version=version,
    author="Mike S Wang",
    author_email="mike.wang@port.ac.uk",
    license="GPLv3",
    description=(
        "Hybrid-basis Fourier analysis of large-scale galaxy clustering."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MikeSWang/Harmonia/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    python_requires='>=3.6',
    project_urls={
        "Documentation": "https://mikeswang.github.io/Harmonia",
        "Source": "https://github.com/MikeSWang/Harmonia/",
    },
)


"""
***************************************************************************
Survey factory (:mod:`~harmonia.surveyor`)
***************************************************************************

Produce from survey definition and specifications the quantities needed
for map analysis, e.g. mask map, selection function, mask/window function
multipoles, covariance estimates and data compressor.

.. note::

    Unless otherwise specified, the length dimension in the module is
    in units of Mpc/:math:`h`.

"""
# No import from ``.synthesis`` to avoid circular import.
from .definition import (
    generate_mask_by_sky_fraction,
    generate_mask_from_map,
    generate_selection_by_cut,
    generate_selection_by_distribution,
    generate_selection_from_samples,
    generate_selection_samples,
)
from .coordinates import (
    cartesian_to_spherical,
    sky_to_spherical,
    spherical_to_cartesian,
    spherical_to_sky,
    to_box_coords,
)



"""
Spherical kernels (:mod:`~harmonia.reader.kernels`)
===========================================================================

Evaluate integration kernels for computing spherical Fourier coupling
coefficients.

"""
import numpy as np

from harmonia.algorithms.bases import spherical_besselj, spherical_harmonic


def angular_kernel(theta, phi, mu, nu, mask=None):
    r"""Evaluate the angular coupling kernel.

    Parameters
    ----------
    theta, phi : float, array_like
        Angular coordinates :math:`\theta`, :math:`\phi`.
    mu, nu : tuple(int, int, int)
        Coefficient triplet index.
    mask : callable or None, optional
        Mask as a function of angular coordinates (default is `None`).

    Returns
    -------
    kernel : complex :class:`numpy.ndarray`
        Angular coupling kernel value.

    """
    kernel = \
        spherical_harmonic(mu[0], mu[1], theta, phi, conj=True) \
        * spherical_harmonic(nu[0], nu[1], theta, phi)

    if callable(mask):
        kernel *= mask(np.column_stack([theta, phi]))

    return kernel


def radial_kernel(r, mu, nu, k_mu, k_nu, selection=None, weight=None,
                  bias_evolution=None, clustering_evolution=None,
                  z_from_r=None, chi_of_z=None):
    """Evaluate the radial coupling kernel.

    Parameters
    ----------
    r : float, array_like
        Radial coordinate.
    mu, nu : tuple(int, int, int)
        Coefficient triplet index.
    k_mu, k_nu : float
        Discrete wave number corresponding to index `mu` or `nu`.
    selection, weight : callable or None, optional
        Selection or weight as a function of the radial coordinate
        (default is `None`).
    bias_evolution, clustering_evolution : callable or None, optional
        Bias and clustering evolution as a function of redshift
        normalised to unity at the current epoch (default is `None`).
    z_from_r : callable or None, optional
        Comoving distance-to-redshift conversion in a variable
        cosmological model (default is `None`).
    chi_of_z : callable or None, optional
        Comoving redshift-to-distance conversion in a fiducial
        cosmological model (default is `None`).

    Returns
    -------
    kernel : float :class:`numpy.ndarray`
        Radial coupling kernel value.

    Raises
    ------
    TypeError
        If `z_from_r` is not callable when any of `bias_evolution`,
        `clustering_evolution` and `chi_of_z` is.

    """
    if any(map(callable, [chi_of_z, clustering_evolution, bias_evolution])) \
            and not callable(z_from_r):
        raise TypeError(
            "`z_from_r` must be callable when any of "
            "`bias_evolution`, `clustering_evolution` and `chi_of_z` is."
        )

    if not callable(chi_of_z):
        r_tilde = r
    else:
        r_tilde = chi_of_z(z_from_r(r))

    kernel = \
        spherical_besselj(mu[0], k_mu * r_tilde) \
        * spherical_besselj(nu[0], k_nu * r)

    if callable(selection):
        kernel *= selection(r)

    if callable(weight):
        kernel *= weight(r_tilde)

    if callable(clustering_evolution):
        kernel *= clustering_evolution(z_from_r(r))

    if callable(bias_evolution):
        kernel *= bias_evolution(z_from_r(r))

    return kernel


def RSD_kernel(r, mu, nu, k_mu, k_nu, selection=None, weight=None,
               weight_derivative=None, growth_evolution=None,
               clustering_evolution=None, z_from_r=None, chi_of_z=None,
               differential_AP_distortion=None):
    """Evaluate the RSD coupling kernel.

    Parameters
    ----------
    r : float, array_like
        Radial coordinate.
    mu, nu : tuple(int, int, int)
        Coefficient triplet index.
    k_mu, k_nu : float
        Discrete wave number corresponding to index `mu` or `nu`.
    selection, weight, weight_derivative : callable or None, optional
        Selection, weight or weight derivative as a function of the
        radial coordinate (default is `None`).
    growth_evolution, clustering_evolution : callable or None, optional
        Growth rate evolution or clustering evolution as a function of
        redshift normalised to unity at the current epoch (default is
        `None`).
    z_from_r : callable or None, optional
        Cosmological comoving distance-to-redshift conversion (default
        is `None`).
    chi_of_z : callable or None, optional
        Fiducial comoving redshift-to-distance conversion (default is
        `None`).
    differential_AP_distortion : callable or None, optional
        Differential AP distortion as a function of redshift (default
        is `None`).

    Returns
    -------
    kernel : float, array_like
        RSD coupling kernel value.

    Raises
    ------
    TypeError
        If `z_from_r` is not callable when any of `growth_evolution`,
        `clustering_evolution` and `chi_of_z` is.
    TypeError
        If `weight` and `weight_derivative` are not both None or callable.
    TypeError
        If `chi_of_z` and `AP_distortion` are not both None or callable.

    """
    if any(map(callable, [chi_of_z, growth_evolution, clustering_evolution])) \
            and not callable(z_from_r):
        raise TypeError(
            "`z_from_r` must be callable when any of "
            "`growth_evolution`, `clustering_evolution` and `chi_of_z` is."
        )

    apply_weight = any(map(callable, [weight, weight_derivative]))
    weight_applicable = all(map(callable, [weight, weight_derivative]))
    if apply_weight and not weight_applicable:
        raise TypeError(
            "`weight` and `weight_derivative` must both be callable "
            "in order to apply weight."
        )

    apply_AP = any(map(callable, [chi_of_z, differential_AP_distortion]))
    AP_applicable = all(map(callable, [chi_of_z, differential_AP_distortion]))
    if apply_AP and not AP_applicable:
        raise TypeError(
            "`chi_of_z` and `differential_AP_distortion` must both be "
            "callable in order to apply AP correction."
        )

    kernel = spherical_besselj(nu[0], k_nu*r, derivative=True)

    if apply_AP:
        r_tilde = chi_of_z(z_from_r(r))
        kernel *= differential_AP_distortion(z_from_r(r))
    else:
        r_tilde = r

    if callable(selection):
        kernel *= selection(r)

    if apply_weight:
        kernel *= (
            weight_derivative(r_tilde)
            * spherical_besselj(mu[0], k_mu * r_tilde) \
            + k_mu * weight(r_tilde) \
            * spherical_besselj(mu[0], k_mu * r_tilde, derivative=True)
        )
    else:
        kernel *= k_mu * spherical_besselj(mu[0], k_mu * r, derivative=True)

    if callable(growth_evolution):
        kernel *= growth_evolution(z_from_r(r))

    if callable(clustering_evolution):
        kernel *= clustering_evolution(z_from_r(r))

    return kernel


def shot_noise_kernel(r, mu, nu, k_mu, k_nu, selection=None, weight=None):
    """Evalaute the two-point correlator shot noise kernel.

    Parameters
    ----------
    r : float, array_like
        Radial coordinate.
    mu, nu : tuple(int, int, int)
        Coefficient triplet index.
    k_mu, k_nu : float
        Discrete wave number corresponding to index `mu` or `nu`.
    selection, weight : callable or None, optional
        Selection or weight as a function of the radial coordinate
        (default is `None`).

    Returns
    -------
    kernel : float, array_like
        Two-point correlator shot noise kernel value.

    """
    kernel = \
        spherical_besselj(mu[0], k_mu * r) \
        * spherical_besselj(nu[0], k_nu * r)

    if callable(selection):
        kernel *= selection(r)

    if callable(weight):
        kernel *= weight(r) ** 2

    return kernel



"""
Coordinates (:mod:`~harmonia.surveyor.coordinates`)
===========================================================================

Handle the survey coordinate systems.

.. autosummary::

    cartesian_to_spherical
    spherical_to_cartesian
    sky_to_spherical
    spherical_to_sky
    to_box_coords

|

"""
# pylint: disable=assignment-from-no-return
import warnings
from functools import wraps

import numpy as np


def cartesian_to_spherical(cartesian_coords):
    r"""Convert 3-d Cartesian coordinate arrays to spherical coordinate
    arrays.

    The coordinate transformation is given by

    .. math::

        r = \sqrt{x^2 + y^2 + z^2} \,, \quad
        \theta = \arccos(z/r) \,, \quad
        \phi = \arctan(y/x) \,,

    where the image of :math:`\arccos` is :math:`[0, \pi]`, and
    :math:`\arctan` has an extended image set :math:`[0, 2\pi]`.

    Parameters
    ----------
    cartesian_coords : float, array_like
        Cartesian coordinates.

    Returns
    -------
    spherical_coords : float :class:`numpy.ndarray`
        Spherical coordinates.

    """
    c_coords = np.atleast_2d(cartesian_coords)
    if not _is_coord_3d(c_coords):
        raise ValueError("`cartesian_coords` is not 3-d.")

    spherical_coords = np.zeros(c_coords.shape)
    spherical_coords[:, 0] = np.linalg.norm(c_coords, axis=1)
    spherical_coords[:, 1] = np.arccos(c_coords[:, 2] / spherical_coords[:, 0])
    spherical_coords[:, 2] = np.mod(
        np.arctan2(c_coords[:, 1], c_coords[:, 0]), 2*np.pi
    )

    return spherical_coords


def spherical_to_cartesian(spherical_coords):
    r"""Convert 3-d spherical coordinate arrays to Cartesian coordinate
    arrays.

    The coordinate transformation is given by

    .. math::

        x = r \sin\theta \cos\phi \,, \quad
        y = r \sin\theta \sin\phi \,, \quad
        z = r \cos\theta \,.

    Parameters
    ----------
    spherical_coords : float, array_like
        Spherical coordinates.

    Returns
    -------
    cartesian_coords : float :class:`numpy.ndarray`
        Cartesian coordinates.

    """
    s_coords = np.atleast_2d(spherical_coords)
    if not _is_coord_3d(s_coords):
        raise ValueError("`spherical_coords` is not 3-d.")

    cartesian_coords = np.zeros(s_coords.shape)
    cartesian_coords[:, 0] = np.sin(s_coords[:, 1]) * np.cos(s_coords[:, 2])
    cartesian_coords[:, 1] = np.sin(s_coords[:, 1]) * np.sin(s_coords[:, 2])
    cartesian_coords[:, 2] = np.cos(s_coords[:, 1])

    cartesian_coords *= (s_coords[:, 0])[:, None]

    return cartesian_coords


def sky_to_spherical(sky_coords, z_to_r=None):
    r"""Convert 3-d (or 2-d) sky coordinate arrays (Z, DEC, RA)
    (or (DEC, RA)) to spherical coordinate arrays.

    The spherical surface coordinate transformation is given by

    .. math::

        \theta = \frac{\pi}{180} (90 - \delta) \,, \quad
        \phi = \frac{\pi}{180} \alpha \,.

    where :math:`\delta` is the declination (DEC) and :math:`\alpha`
    the right ascension (RA) both given in degrees.

    The radial coordinate transformation from redshifts is given by
    :meth:`cosmo.comoving_distance` if the input coordinates are 3-d
    and a redshift-to-distance conversion function is provided.

    Parameters
    ----------
    sky_coords : float, array_like
        Sky coordinates (2-d or 3-d).
    z_to_r : callable or None, optional
        Redshift-to-distance conversion (default is `None`).

    Returns
    -------
    spherical_coords : float :class:`numpy.ndarray`
        Spherical surface coordinates.

    """
    sky_coords = np.atleast_2d(sky_coords)

    if _is_coord_3d(sky_coords):
        if not callable(z_to_r):
            raise TypeError(
                "`z_to_r` must be provided for radial "
                "coordinate transformation."
            )

        spherical_coords = np.zeros(sky_coords.shape)
        spherical_coords[:, 0] = z_to_r(sky_coords[:, 0])
        spherical_coords[:, 1] = np.deg2rad(90 - sky_coords[:, 1])
        spherical_coords[:, 2] = np.deg2rad(sky_coords[:, 2])

        return spherical_coords

    if _is_coord_2d(sky_coords):

        spherical_coords = np.zeros(sky_coords.shape)
        spherical_coords[:, 0] = np.deg2rad(90 - sky_coords[:, 0])
        spherical_coords[:, 1] = np.deg2rad(sky_coords[:, 1])

        return spherical_coords

    raise ValueError("Check dimensions of input `sky_coords`.")


def spherical_to_sky(spherical_coords, z_from_r=None):
    r"""Convert 3-d spherical coordinate arrays to sky coordinate arrays.

    The spherical surface coordinate transformation is given by

    .. math::

        \delta = 90 - \frac{180}{\pi} \theta \,, \quad
        \alpha = \frac{180}{\pi} \phi \,.

    where :math:`\delta` is the declination (DEC) and :math:`\alpha`
    the right ascension (RA) both given in degrees.

    The radial coordinate transformation is given by a
    distance-to-redshift conversion if it provided.

    Parameters
    ----------
    sky_coords : float, array_like
        Sky coordinates.
    z_from_r : callable or None, optional
        Distance-to-redshift conversion (default is `None`).

    Returns
    -------
    sky_coords : float :class:`numpy.ndarray`
        Sky coordinates.

    """
    spherical_coords = np.atleast_2d(spherical_coords)
    if not _is_coord_3d(spherical_coords):
        raise ValueError("`spherical_coords` is not 3-d.")

    sky_coords = np.zeros(spherical_coords.shape)

    sky_coords[:, 1] = 90 - np.rad2deg(spherical_coords[:, 1])
    sky_coords[:, 2] = np.rad2deg(spherical_coords[:, 2])

    if z_from_r is None:

        sky_coords = np.delete(sky_coords, 0, axis=1)

        return sky_coords

    sky_coords[:, 0] = z_from_r(spherical_coords[:, 0])

    return sky_coords


def _is_coord_2d(coords):

    try:
        return np.size(coords, axis=-1) == 2
    except IndexError: # scalars
        return False


def _is_coord_3d(coords):

    try:
        return np.size(coords, axis=-1) == 3
    except IndexError: # scalars
        return False


# pylint: disable=unused-argument
def to_box_coords(native_coord_system, box_centre=None,
                  conversion_kwargs=None):
    """Convert a function defined for a native 3-d curvilinear coordinate
    system to an equivalent function defined for a box in
    Cartesian coordinates.

    Parameters
    ----------
    native_coord_system : {'null', 'spherical', 'sky'}, optional
        Native coordinate system of the function.  If 'cartesian' (with
        the implicit assumption that the origin is at the box centre),
        the function is unconverted unless `box_shift` is also provided.
    box_centre : float, array_like or Nonw
        If provided, translate the box coordinates so that the (positive)
        `box_centre` is moved to(0, 0, 0).  This is needed to when the
        wrapped function needs to accept box coordinates with the origin
        placed at its corner whilst the `native_coord_system` has the
        origin at the centre of the box (e.g. 'spherical').
    conversion_kwargs : dict or None, optional
        Additional parameters to use in conversion, e.g. `z_from_r`
        if `native_coords` is 'sky' (default is `None`).

    Returns
    -------
    callable
        The original function now accepting Cartesian coordinates.

    """
    def decorator(func):
        @wraps(func)
        def wrapper(*coord_arrs, **kwargs):

            nonlocal native_coord_system

            # Offer the option to do nothing.
            if box_centre is None and native_coord_system == 'null':
                return func(*coord_arrs, **kwargs)

            native_coord_system = 'cartesian' \
                if native_coord_system == 'null' \
                else native_coord_system

            if len(coord_arrs) == 1:
                unpacked = False
            elif len(coord_arrs) == 3:
                unpacked = True
            else:
                raise ValueError(
                    "Wrapped coordinate function must only have positional "
                    "arguments that are 3-d coordinate arrays."
                )

            if unpacked:
                coords = np.column_stack(coord_arrs)
            else:
                coords, = coord_arrs

            if box_centre is None:
                cart_coords = np.atleast_2d(coords)
            else:
                if np.any(np.sign(box_centre) < 0):
                    warnings.warn(
                        "`box_shift` is usually positive to ensure "
                        "the origin of the Cartesian coordinates "
                        "is at a corner of the box."
                    )
                cart_coords = np.subtract(coords, box_centre)

            if native_coord_system == 'cartesian':
                native_coords = cart_coords
            elif native_coord_system == 'spherical':
                native_coords = cartesian_to_spherical(cart_coords)
            elif native_coord_system == 'sky':
                native_coords = spherical_to_sky(
                    cartesian_to_spherical(cart_coords), **conversion_kwargs
                )
            else:
                raise ValueError(
                    "Unsupported `native_coord_system`: "
                    f"{native_coord_system}."
                )

            if unpacked:
                native_coord_arrs = np.hsplit(native_coords, 3)
            else:
                native_coord_arrs = (native_coords,)

            return func(*native_coord_arrs, **kwargs)
        return wrapper
    return decorator



"""
Fourier bases (:mod:`~harmonia.algorithms.bases`)
===========================================================================

Compute quantities related to Fourier basis functions.

.. autosummary::

    spherical_harmonic
    spherical_besselj
    spherical_besselj_root

|

"""
# pylint: disable=no-name-in-module
import numpy as np
from mpmath import besseljzero
from scipy.special import sph_harm, spherical_jn

from harmonia.utils import binary_search


def spherical_harmonic(ell, m, theta, phi, conj=False):
    r"""Evaluate the spherical harmonic function.

    This returns :math:`Y_{\ell m}(\theta, \phi)` of degree :math:`\ell
    \geqslant 0` and order :math:`\vert {m} \vert \leqslant \ell` at the
    polar angle :math:`\theta \in [0, \pi]` and the azimuthal angle
    :math:`\phi \in [0, 2\pi]`.

    Parameters
    ----------
    ell : int, array_like
        Degree of the spherical harmonic function, ``ell >= 0``.
    m : int, array_like
        Order of the spherical harmonic function, ``|m| <= ell``.
    theta : float, array_like
        Polar angle in the interval ``[0, pi]``.
    phi: float, array_like
        Azimuthal angle in the interval ``[0, 2*pi]``.
    conj : bool, optional
        If `True` (default is `False`), return the complex conjugate.

    Returns
    -------
    complex, array_like
        :math:`Y_{\ell m}` value at `theta` and `phi`.

    """
    return np.conj(sph_harm(m, ell, phi, theta)) if conj \
        else sph_harm(m, ell, phi, theta)


def spherical_besselj(ell, x, derivative=False):
    r"""Evaluate the spherical Bessel function of the first kind or
    its derivative.

    This returns :math:`j_{\ell}(x)` or :math:`j'_{\ell}(x)` of order
    :math:`\ell \geqslant 0`.

    Parameters
    ----------
    ell : int, array_like
        Order of the spherical harmonic function, ``ell >= 0``.
    x : float, array_like
        Argument of the spherical Bessel function.
    derivative : bool, optional
        If `True` (default is `False`), evaluate the derivative instead.

    Returns
    -------
    float, array_like
        :math:`j_{\ell}` or :math:`j'_{\ell}` value at `x`.

    """
    return spherical_jn(ell, x, derivative=derivative)


def spherical_besselj_root(ell, nmax, only=True, derivative=False):
    r"""Find positive zero(s) :math:`u_{\ell n}` of the spherical Bessel
    function :math:`j_{\ell}` of the first kind of order :math:`\ell`, or
    its derivative :math:`j'_{\ell}`, up to a maximum number
    :math:`n_\textrm{max}`.

    Solving for roots of the spherical Bessel function relies on the
    identity :math:`j_\ell(x) = \sqrt{\pi/(2x)} J_{\ell + 1/2}(x)`, where
    :math:`J_\ell(x)` is the Bessel funcion of the first kind.  Solving for
    roots of the derivative function employs the interval bisection method,
    with the interval ansatz :math:`\ell + 1 \leqslant x \leqslant
    n_\textrm{max} \operatorname{max}\{4, \ell\}`.

    Parameters
    ----------
    ell : int
        Order of the spherical Bessel function, ``ell >= 0``.
    nmax : int
        Maximum number of positive zeros to be found, ``nmax >= 1``.
    only : bool, optional
        If `True` (default), return the maximal root only.
    derivative : bool, optional
        If `True` (default is `False`), compute the zero(s) of the
        derivative function instead.

    Returns
    -------
    u_ell : float, array_like
        Positive zero(s) for order `ell` (in ascending order).

    """
    if not derivative:
        # `mpmath` returns a ``mpf`` float which needs conversion.
        if only:
            u_ell = float(besseljzero(ell + 0.5, nmax, derivative=0))
        else:
            u_ell = np.asarray([
                float(besseljzero(ell + 0.5, n, derivative=0))
                for n in range(1, nmax + 1)
            ])
    else:
        u_elln_list = binary_search(
            lambda x: spherical_besselj(ell, x, derivative=True),
            ell + 1, nmax * max(4, ell),
            maxnum=nmax
        )
        u_ell = u_elln_list[-1] if only else u_elln_list

    return u_ell



import pytest

from harmonia.surveyor.definition import (
    generate_mask_by_sky_fraction,
    generate_mask_from_map,
    generate_selection_by_distribution,
    generate_selection_from_samples,
    generate_selection_samples,
)


@pytest.mark.skip(reason="Unimplemented.")
def test_generate_mask_by_sky_fraction():
    print("Unimplemented test.")
    raise NotImplementedError


@pytest.mark.skip(reason="Unimplemented.")
def test_generate_mask_from_map():
    print("Unimplemented test.")
    raise NotImplementedError


@pytest.mark.skip(reason="Unimplemented.")
def test_generate_selection_by_distribution():
    print("Unimplemented test.")
    raise NotImplementedError


@pytest.mark.skip(reason="Unimplemented.")
def test_generate_selection_from_samples():
    print("Unimplemented test.")
    raise NotImplementedError


@pytest.mark.skip(reason="Unimplemented.")
def test_generate_selection_samples():
    print("Unimplemented test.")
    raise NotImplementedError
